/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   Benjamin Lerman (benjamin.lerman@ens.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File conndata.c */

/* Thus speaketh ye mighty CVS:
 * $Id: conndata.c,v 1.1 1999/05/01 22:46:12 david Exp $ */

/* This file contains the declarations for server.c */

/* Include files */

#include <stdio.h>		/* Standard I/O */

#include <errno.h>		/* Error numbers */
#include <string.h>		/* String functions such as strerror() */

#include <sys/socket.h>		/* Socket types */

#include <netinet/in.h>		/* INET domain (TCP/IP) sockets */

#include <pthread.h>		/* Posix threads */


#include "server.h"

/* Data structures - see comments in server.h */

int gsock;

int lsock_nb = 0;
struct lsock_s **lsock_tab = NULL;

pthread_mutex_t lsock_mutex = PTHREAD_MUTEX_INITIALIZER;

long conn_count = 0;

pthread_cond_t handler_wake_cond = PTHREAD_COND_INITIALIZER;
pthread_mutex_t handler_wake_mutex = PTHREAD_MUTEX_INITIALIZER;
long handler_wake_count = 0;

/* Useful functions */

/* ``Safe'' pthread functions.  I write those because I can't affort
 * to write the error checking clause *every* damn time I lock or
 * unlock a mutex. */

inline void
safe_pthread_mutex_init(pthread_mutex_t *mutex,
			const pthread_mutexattr_t *mutexattr)
{
  int error;
  if ((error=pthread_mutex_init(mutex,mutexattr)))
    {
      fprintf(stderr,"Initializing mutex: %s\n",strerror(error));
      exit(1);
    }
}

inline void
safe_pthread_mutex_lock(pthread_mutex_t *mutex)
{
  int error;
  if ((error=pthread_mutex_lock(mutex)))
    {
      fprintf(stderr,"Locking mutex: %s\n",strerror(error));
      exit(1);
    }
}

inline int
safe_pthread_mutex_trylock(pthread_mutex_t *mutex)
{
  int error;
  if ((error=pthread_mutex_trylock(mutex)))
    {
      if (error==EBUSY) return EBUSY;
      fprintf(stderr,"Locking mutex: %s\n",strerror(error));
      exit(1);
    }
  return 0;
}

inline void
safe_pthread_mutex_unlock(pthread_mutex_t *mutex)
{
  int error;
  if ((error=pthread_mutex_unlock(mutex)))
    {
      fprintf(stderr,"Unlocking mutex: %s\n",strerror(error));
      exit(1);
    }
}
