/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Antoine Allain (allain@aix.ensam.fr)
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   Benjamin Lerman (benjamin.lerman@ens.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File handler.c */

/* Thus speaketh ye mighty CVS:
 * $Id: handler.c,v 1.9 1999/05/29 18:15:06 david Exp $ */

/* This program contains the main function of the handler threads. */

/* CVS identity of this file. */
static char id[] = "$Id: handler.c,v 1.9 1999/05/29 18:15:06 david Exp $";

/* Include files */

#ifndef _THREAD_SAFE
#define _THREAD_SAFE		/* Make sure libc is reentrant. */
#endif

#include <stdio.h>		/* Standard I/O */
#include <stdlib.h>		/* Standard library functions */

#include <unistd.h>		/* UNIX standard functions */

#include <pthread.h>		/* Posix threads */


#include "server.h"

#include "lgdproto.h"

#include "map.h"

/* Main functions */

void
reply_error (struct lsock_s *sck, unsigned long callback,
	     unsigned long errcode)
     /* Return an error code. */
{
  char data[20];

  if ( ! sck->sopen ) return;
  *(unsigned long *)(data) = htonl (LGD_PROTO_REP_ERR);
  *(unsigned long *)(data+4) = htonl (20);
  *(unsigned long *)(data+8) = htonl (0); /* Reserved */
  *(unsigned long *)(data+12) = callback;
  *(unsigned long *)(data+16) = htonl (errcode);
  if ( write (sck->s, data, sizeof(data)) < 0 )
    {
      perror ("write()ing to socket");
      exit (1);
    }
}

void
reply_nothing (struct lsock_s *sck, unsigned long callback)
     /* Make a no-content response. */
{
  char data[16];

  if ( ! sck->sopen ) return;
  *(unsigned long *)(data) = htonl (LGD_PROTO_REP_SYN);
  *(unsigned long *)(data+4) = htonl (16);
  *(unsigned long *)(data+8) = htonl (0); /* Reserved */
  *(unsigned long *)(data+12) = callback;
  if ( write (sck->s, data, sizeof(data)) < 0 )
    {
      perror ("write()ing to socket");
      exit (1);
    }
}

void
reply_simple (struct lsock_s *sck, unsigned long callback,
	      unsigned long reply)
     /* Return just a 32-bit quantity. */
{
  char data[20];

  if ( ! sck->sopen ) return;
  *(unsigned long *)(data) = htonl (LGD_PROTO_REP_SYN);
  *(unsigned long *)(data+4) = htonl (20);
  *(unsigned long *)(data+8) = htonl (0); /* Reserved */
  *(unsigned long *)(data+12) = callback;
  *(unsigned long *)(data+16) = htonl (reply);
  if ( write (sck->s, data, sizeof(data)) < 0 )
    {
      perror ("write()ing to socket");
      exit (1);
    }
}

void
reply_double (struct lsock_s *sck, unsigned long callback,
	      unsigned long reply1, unsigned long reply2)
     /* Return two 32-bit quantities. */
{
  char data[24];

  if ( ! sck->sopen ) return;
  *(unsigned long *)(data) = htonl (LGD_PROTO_REP_SYN);
  *(unsigned long *)(data+4) = htonl (24);
  *(unsigned long *)(data+8) = htonl (0); /* Reserved */
  *(unsigned long *)(data+12) = callback;
  *(unsigned long *)(data+16) = htonl (reply1);
  *(unsigned long *)(data+20) = htonl (reply2);
  if ( write (sck->s, data, sizeof(data)) < 0 )
    {
      perror ("write()ing to socket");
      exit (1);
    }
}

void
do_greet (struct lsock_s *sck)
{
  char data[24];

  sck->state = CONN_STATE_NORMAL;
  if ( ! sck->sopen ) return;
  *(unsigned long *)(data) = htonl (LGD_PROTO_REP_CON);
  *(unsigned long *)(data+4) = htonl (24);
  *(unsigned long *)(data+8) = htonl (0); /* Reserved */
  *(unsigned long *)(data+12) = htonl (LGD_PROTO_MAGIC);
  *(unsigned long *)(data+16) = htonl (LGD_PROTO_VERNUM);
  *(unsigned long *)(data+20) = htonl (1);
  if ( write (sck->s, data, sizeof(data)) < 0 )
    {
      perror ("write()ing to socket");
      exit (1);
    }
}

void
do_request (struct lsock_s *sck)
     /* Handle a request on the given connection. */
{
  unsigned long len,req,callback;
#if ( DEBUG_LEVEL >= 3 )
  fprintf (stderr, "[%0#10lx]Considering request.\n",
	   pthread_self ());
#endif /* DEBUG_LEVEL >= 3 */
  if ( sck->buflen < 8 )
    /* We don't have the request code and length yet. */
    {
      if ( ! sck->sopen )
	/* Socket is closed so new data isn't likely to arrive. */
	{
	  sck->state = CONN_STATE_DEAD;
	  sck->handwork = 1; /* Make sure we loop again. */
	}
      else
	sck->handwork = 0;
      return;
    }
  req = ntohl (*(unsigned long *)(sck->buffer));
  len = ntohl (*(unsigned long *)(sck->buffer+4));
  if ( sck->buflen >= 16 )
    callback = *(unsigned long *)(sck->buffer+12);
  else
    callback = htonl (0); /* Yes, this is htonl() and not ntohl(). */
  if ( len == 0 )
    /* XXX - What should we do in this case?  I don't know.  Let's
     * just silently close the connection.  Note that we allow request
     * lengths less than 16, which have pretty weird effects... */
    {
      if ( sck->sopen )
	shutdown (sck->s, 0);
      sck->state = CONN_STATE_DEAD;
      sck->buflen = 0;
      sck->handwork = 1; /* Make sure we loop again. */
      return;
    }
  if ( len > BUFSIZE )
    /* Request is too long to fit in buffer no matter what.
     * Complain. */
    {
#if ( DEBUG_LEVEL >= 1 )
      fprintf (stderr, "[%0#10lx]Request is too long.\n",
	       pthread_self ());
#endif /* DEBUG_LEVEL >= 1 */
      reply_error (sck, callback, LGD_PROTO_ERR_TOOLONG);
      if ( sck->sopen )
       	shutdown (sck->s, 0);
      /* Someone else will take care of closing s. */
      sck->state = CONN_STATE_DEAD;
      sck->buflen = 0;
      sck->handwork = 1; /* Make sure we loop again. */
      /* XXX - Think of something smarter. */
      return;
    }
  if ( sck->buflen < len )
    /* We don't have the full request yet. */
    {
      if ( ! sck->sopen )
	/* Socket is closed so new data isn't likely to arrive. */
	{
	  len = sck->buflen;
	  /* Redo with what we have. */
	  return; /* XXX - Whatzat?  Didn't we say ``redo''? */
	}
      sck->handwork = 0;
      return;
    }
#if ( DEBUG_LEVEL >= 2 )
  fprintf (stderr, "[%0#10lx]Now handling request: req=%0#10lx, len=%0#10lx\n",
	   pthread_self (), req, len);
#endif /* DEBUG_LEVEL >= 2 */
  switch ( req )
    {
    case LGD_PROTO_REQ_NOP:
      /* NOP */
      reply_nothing (sck, callback);
      break;
    case LGD_PROTO_REQ_ACK:
      /* ACK? */
      reply_simple (sck, callback, 1);
      break;
    case LGD_PROTO_REQ_MAGIC:
      /* MAGIC */
      reply_simple (sck, callback, LGD_PROTO_MAGIC);
      break;
    case LGD_PROTO_REQ_VERNUM:
      /* VERNUM */
      reply_simple (sck, callback, LGD_PROTO_VERNUM);
      break;
    case LGD_PROTO_REQ_DIE:
      /* DIE */
      reply_nothing (sck, callback);
      exit (1);
      break;
    case LGD_PROTO_REQ_IDENT:
      /* IDENT */
      if ( ( len >= 16 )
	   && ( strncmp (sck->buffer+16, "legendes", len-16) == 0 ) )
	reply_simple (sck, callback, 1);
      else
	reply_error (sck, callback, LGD_PROTO_ERR_REFUSED);
      break;
    case LGD_PROTO_REQ_MAPQ:
      /* MAPQ */
      {
	unsigned long x, y;
	if ( len >= 20 )
	  x = ntohl (*(unsigned long *)(sck->buffer+16));
	else
	  x = 0;
	if ( len >= 24 )
	  y = ntohl (*(unsigned long *)(sck->buffer+20));
	else
	  y = 0;
	if ( ( x >= 0 ) && ( y >= 0 )
	     && ( x < MAP_SIZE_X ) && ( y < MAP_SIZE_Y ) )
	  reply_simple (sck, callback, read_map (x, y));
	else
	  reply_error (sck, callback, LGD_PROTO_ERR_OBOUNDS);
      }
      break;
    case LGD_PROTO_REQ_WHERE:
      /* WHERE */
      reply_double (sck, callback, sck->x, sck->y);
      break;
    default:
      reply_error (sck, callback, LGD_PROTO_ERR_UNKNOWN);
      break;
    }
  memmove (sck->buffer, sck->buffer+len, len);
  sck->buflen -= len;
  /* *Don't* set sck->handwork back to 0, because there may be another
   * request in the buffer. */
#if ( DEBUG_LEVEL >= 2 )
  fprintf (stderr, "[%0#10lx]Request handled.\n",
	   pthread_self ());
#endif /* DEBUG_LEVEL >= 2 */
}

void
handler_loop (struct lsock_s *sck)
     /* This function performs a loop that handles the requests that
      * the listener thread has put in the connection buffer. */
{
  safe_pthread_mutex_lock (&sck->d_mutex);
  while (1)
    {
      while ( ! sck->handwork )
	pthread_cond_wait (&sck->h_cond, &sck->d_mutex);
      /* Do everything */
      switch ( sck->state )
	{
	case CONN_STATE_NEW:
	  do_greet (sck);
	  break;
	case CONN_STATE_NORMAL:
	  do_request (sck);
	  break;
	case CONN_STATE_DEAD:
	  /* The time has come to die now */
	  safe_pthread_mutex_unlock (&sck->d_mutex);
	  pthread_exit (NULL);
	  break;
	}
    }
}
